import asyncio
import json
import signal
import sys
import time
import os
import shutil
import random
from unittest import IsolatedAsyncioTestCase

sys.path.append('../')
from kafka_python_app.connector import ListenerConfig, KafkaConnector, ConsumerRecord
from loguru_logger_lite import Logger


KILL = False


class GracefulKiller:
    obj = None

    def exit_gracefully(self, *args):
        global KILL
        KILL = True
        if self.obj:
            self.obj.close()


killer = GracefulKiller()
signal.signal(signal.SIGINT, killer.exit_gracefully)
signal.signal(signal.SIGTERM, killer.exit_gracefully)

msg_counter = 0
time_start = 0


def process_message(message: ConsumerRecord) -> None:
    global msg_counter
    msg = json.loads(message.value)
    print(f'Received: {msg}: n={msg_counter}: after: {round(time.time() - time_start, 3)}s')
    lines = [f'Test line {i}\n' for i in range(200000)]
    with open(f'temp_data/test_{msg_counter}.txt', 'w') as f:
        f.writelines(lines)

    time.sleep(msg['sleep_time']/1000)
    msg_counter += 1


async def process_message_async(message: ConsumerRecord) -> None:
    global msg_counter
    msg = json.loads(message.value)
    print(f'Received: {msg}: n={msg_counter}: after: {round(time.time() - time_start, 3)}s')
    lines = [f'Test line {i}\n' for i in range(200000)]
    with open(f'temp_data/test_{msg_counter}.txt', 'w') as f:
        f.writelines(lines)
    msg_counter += 1
    for i in range(10):
        await asyncio.sleep(msg['sleep_time']/1000)


class TestKafkaListener(IsolatedAsyncioTestCase):
    LOGGER = Logger.get_default_logger()
    KAFKA_BOOTSTRAP_SERVERS = ['127.0.0.1:9092']
    TEST_TOPIC = 'test_topic'

    kafka_listener_config = ListenerConfig(
        bootstrap_servers=KAFKA_BOOTSTRAP_SERVERS,
        process_message_cb=process_message_async,
        consumer_config={
            'group_id': 'test_group',
            # "max_poll_records": 50
        },
        topics=[TEST_TOPIC],
        logger=LOGGER
    )

    producer = KafkaConnector.get_producer(
        KAFKA_BOOTSTRAP_SERVERS,
        # {'batch_size': 10}
    )
    listener = KafkaConnector.get_listener(kafka_listener_config)

    killer.obj = listener

    if os.path.exists('temp_data'):
        shutil.rmtree('temp_data')
    os.makedirs('temp_data')

    async def watch_counter(self):
        global time_start
        while True:
            if msg_counter == 100:
                self.LOGGER.info('Received {} message and now closing...'.format(msg_counter))
                self.listener.stop = True
                break
            if KILL:
                break
            await asyncio.sleep(1)
        print(f'Total time: {time.time() - time_start}')

    async def test_listen(self):
        global time_start
        time_start = time.time()
        for i in range(100):
            sleep_time = random.randint(100, 800)
            msg = json.dumps({'sleep_time': sleep_time})
            self.producer.send(self.TEST_TOPIC, msg)
        await asyncio.sleep(0.5)
        self.producer.close()
        await asyncio.gather(self.listener.listen(), self.watch_counter())



import enum
import sys
import asyncio
import signal
import json
from typing import Any
import pydantic
from unittest import IsolatedAsyncioTestCase

sys.path.append('../')
from kafka_python_app.app import AppConfig, KafkaApp
from kafka_python_app.connector import KafkaConnector
from loguru_logger_lite import Logger


KILL = False


class GracefulKiller:
    obj = None

    def exit_gracefully(self, *args):
        global KILL
        KILL = True
        if self.obj:
            self.obj.close()


class Events(enum.Enum):
    PROCESS_PERSON = 'process_person'
    PROCESS_COMPANY = 'process_company'
    PROCESS_PERSON_ASYNC = 'process_person_async'
    PROCESS_COMPANY_ASYNC = 'process_company_async'


class PersonPayload(pydantic.BaseModel):
    firs_name: str
    last_name: str
    age: int


class CompanyPayload(pydantic.BaseModel):
    name: str
    stock_value: float


class Message(pydantic.BaseModel):
    event: str
    payload: Any


killer = GracefulKiller()
signal.signal(signal.SIGINT, killer.exit_gracefully)
signal.signal(signal.SIGTERM, killer.exit_gracefully)

msg_counter = 0


class TestKafkaApp(IsolatedAsyncioTestCase):
    LOGGER = Logger.get_default_logger()
    KAFKA_BOOTSTRAP_SERVERS = ['127.0.0.1:9092']
    TEST_TOPIC = 'test_topic'

    config = AppConfig(
        app_name='Test application',
        bootstrap_servers=KAFKA_BOOTSTRAP_SERVERS,
        consumer_config={
            'group_id': 'test_app_group'
        },
        listen_topics=[TEST_TOPIC],
        logger=LOGGER
    )

    producer = KafkaConnector.get_producer(KAFKA_BOOTSTRAP_SERVERS)
    app = KafkaApp(config)
    killer.obj = app

    async def watch_counter(self):
        while True:
            if msg_counter >= 2:
                self.LOGGER.info('Received {} message and now closing...'.format(msg_counter))
                self.app.close()
                break
            if KILL:
                break
            await asyncio.sleep(0.1)
        self.assertEqual(msg_counter, 2)

    async def test_handle(self):
        @self.app.on(Events.PROCESS_PERSON.value, topic=self.TEST_TOPIC)
        def handle_person(message, **kwargs):
            global msg_counter
            print('Handling "process_person" event..')
            print('Received: {}\n'.format(message))
            msg_counter += 1

        @self.app.on(Events.PROCESS_COMPANY.value, topic=self.TEST_TOPIC)
        def handle_company(message, **kwargs):
            global msg_counter
            print('Handling "process_company" event..')
            print('Received: {}\n'.format(message))
            msg_counter += 1

        @self.app.on(Events.PROCESS_PERSON_ASYNC.value, topic='other')
        async def handle_person(message, **kwargs):
            global msg_counter
            print('Handling "process_person_async" event..')
            print('Received: {}\n'.format(message))
            msg_counter += 1

        @self.app.on(Events.PROCESS_COMPANY_ASYNC.value, topic='other')
        async def handle_company(message, **kwargs):
            global msg_counter
            print('Handling "process_company_async" event..')
            print('Received: {}\n'.format(message))
            msg_counter += 1

        messages = [
            {
                'event': Events.PROCESS_PERSON.value,
                'payload': {
                    'first_name': 'John',
                    'last_name': 'Doe',
                    'age': 35
                }
            },
            {
                'event': Events.PROCESS_COMPANY.value,
                'payload': {
                    'name': 'SomeCompany',
                    'stock_value': 1224.55
                }
            },
            {
                'event': Events.PROCESS_PERSON_ASYNC.value,
                'payload': {
                    'first_name': 'John Async',
                    'last_name': 'Doe',
                    'age': 15
                }
            },
            {
                'event': Events.PROCESS_COMPANY_ASYNC.value,
                'payload': {
                    'name': 'SomeCompany Async',
                    'stock_value': 12424.55
                }
            }
        ]

        for msg_obj in messages:
            msg = Message(**msg_obj)
            await asyncio.sleep(0.5)
            self.producer.send(self.TEST_TOPIC, json.loads(msg.json(exclude_unset=True)))

        await asyncio.sleep(0.1)
        self.producer.close()
        await asyncio.gather(self.app.run(), self.watch_counter())




import asyncio
import inspect
import time
import uuid

from kafka import KafkaProducer, KafkaConsumer, ConsumerRebalanceListener
from kafka.consumer.fetcher import ConsumerRecord
import pydantic
from typing import List, Dict, Optional, Callable, Any, Tuple, Union, Coroutine
import json
import sys
import logging


class ListenerConfig(pydantic.BaseModel):
    bootstrap_servers: List[str]
    process_message_cb: Union[Callable[[ConsumerRecord], None], Callable[[ConsumerRecord], Coroutine[Any, Any, None]]]
    consumer_config: Optional[Dict]
    topics: List[str]
    logger: Optional[Any]


class ProducerRecord(pydantic.BaseModel):
    value: Any
    key: Optional[Any]
    headers: Optional[List[Tuple[str, bytes]]]
    partition: Optional[int]
    timestamp_ms: Optional[int]


class KafkaConnector:

    @staticmethod
    def get_producer(bootstrap_servers, producer_config: Dict = None):
        config = {
            'bootstrap_servers': bootstrap_servers,
            'key_serializer': lambda x: x.encode('utf-8') if x else x,
            'value_serializer': lambda x: json.dumps(x).encode('utf-8'),
        }
        if producer_config:
            # for key in producer_config.keys():
            #     config[key] = producer_config.get(key)
            config = {**config, **producer_config}
        return KafkaProducer(**config)

    @staticmethod
    def get_listener(config: ListenerConfig):
        return KafkaListener(config)


# class MyConsumerRebalanceListener(ConsumerRebalanceListener):
#     def __init__(self, logger: loguru_logger = None):
#         super(MyConsumerRebalanceListener, self).__init__()
#         if logger is None:
#             _logger_cls = Logger()
#             self.logger = _logger_cls.default_logger
#         else:
#             self.logger = logger
#
#     def on_partitions_assigned(self, assigned):
#         self.logger.info('Partitions assigned: {}'.format(assigned))
#
#     def on_partitions_revoked(self, revoked):
#         self.logger.info('Partitions revoked: {}'.format(revoked))


class KafkaListener:

    def __init__(self, config: ListenerConfig):
        self.stop = False
        self.config = config
        if not self.config.logger:
            logging.basicConfig()
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
        else:
            self.logger = self.config.logger

        config = {
            'bootstrap_servers': config.bootstrap_servers,
            'group_id': str(uuid.uuid4()),
            'auto_offset_reset': 'earliest',
            'enable_auto_commit': False,
            'key_deserializer': lambda x: x.decode('utf-8') if x else x,
            'value_deserializer': lambda x: json.loads(x.decode('utf-8')),
            'session_timeout_ms': 25000
        }
        if self.config.consumer_config:
            # for key in self.config.consumer_config.keys():
            #     config[key] = self.config.consumer_config.get(key)
            config = {**config, **self.config.consumer_config}

        self.consumer = KafkaConsumer(**config)
        self.consumer.subscribe(topics=self.config.topics)

    # def __del__(self):
    #     self.stop = True
    #     self.close()

    def close(self):
        # self.logger.info('Closing listener..')
        self.consumer.close()
        self.logger.info('Kafka listener closed.')

    async def listen(self):
        # try:
        while True:
            message_batch = self.consumer.poll()
            time_up = time.time()
            if inspect.iscoroutinefunction(self.config.process_message_cb):
                try:
                    await asyncio.gather(*[
                        self.config.process_message_cb(message) for partition_batch in message_batch.values()
                        for message in partition_batch
                    ])
                except Exception as e:
                    self.logger.error('Exception: type: {} line#: {} msg: {}'.format(sys.exc_info()[0],
                                                                                     sys.exc_info()[
                                                                                         2].tb_lineno,
                                                                                     str(e)))
            else:
                for partition_batch in message_batch.values():
                    for message in partition_batch:
                        try:
                            self.config.process_message_cb(message)
                        except Exception as e:
                            self.logger.error('Exception: type: {} line#: {} msg: {}'.format(sys.exc_info()[0],
                                                                                             sys.exc_info()[
                                                                                                 2].tb_lineno,
                                                                                             str(e)))
            self.logger.trace(f'Batch processed in {time.time() - time_up}s')
            try:
                self.consumer.commit()
            except Exception as e:
                self.logger.error('Exception: type: {} line#: {} msg: {}'.format(sys.exc_info()[0],
                                                                                 sys.exc_info()[
                                                                                     2].tb_lineno,
                                                                                 str(e)))

            if self.stop:
                # self.logger.info('Shutting down..')
                break

            await asyncio.sleep(0.001)

        # except Exception as e:
        #     self.logger.error('Exception: type: {} line#: {} msg: {}'.format(sys.exc_info()[0],
        #                                                                      sys.exc_info()[2].tb_lineno,
        #                                                                      str(e)))
        #
        # finally:
        #     self.close()



import inspect
import json
import sys
import uuid
from typing import Dict, List, Optional, Any, Callable, Union, Coroutine, Tuple
import pydantic
import logging
import hashlib
import time
import asyncio
from collections import deque
from kafka_python_app.connector import ListenerConfig, KafkaConnector, ConsumerRecord, ProducerRecord


def _get_event_id_hash(event_id: str, service_id: str):
    _id = '.'.join([service_id, event_id])
    return hashlib.sha256(_id.encode('utf-8')).hexdigest()


class TransactionPipeWithReturnOptions(pydantic.BaseModel):
    response_event_name: str
    response_from_topic: str
    cache_client: Any
    return_event_timeout: int


class TransactionPipeResultOptions(pydantic.BaseModel):
    pipe_event_name: str
    pipe_to_topic: str
    with_response_options: Optional[TransactionPipeWithReturnOptions]


class MessageTransaction(pydantic.BaseModel):
    fnc: Callable
    args: Optional[Dict]
    pipe_result_options: Optional[Union[Callable, TransactionPipeResultOptions]]


class MessagePipeline(pydantic.BaseModel):
    transactions: List[MessageTransaction]
    app_id: Optional[str]
    logger: Optional[Any]

    __exceptions: List[Any] = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not kwargs.get('logger'):
            logging.basicConfig()
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)

    async def execute(self, message, emitter, message_key_as_event, **kwargs):
        try:
            for tx in self.transactions:
                self.logger.debug(f'Executing transaction: '
                                  f'name: {getattr(tx.fnc, "__name__", repr(tx.fnc))}; '
                                  f'args: {tx.args}')
                if inspect.iscoroutinefunction(tx.fnc):
                    if tx.args:
                        message = await tx.fnc(message, self.logger, **{**tx.args, **kwargs})
                    else:
                        message = await tx.fnc(message, self.logger, **kwargs)
                else:
                    if tx.args:
                        message = tx.fnc(message, self.logger, **{**tx.args, **kwargs})
                    else:
                        message = tx.fnc(message, self.logger, **kwargs)

                if tx.pipe_result_options:
                    if inspect.isfunction(tx.pipe_result_options):
                        pipe_result_options: TransactionPipeResultOptions = tx.pipe_result_options(message, self.logger,
                                                                                                   **kwargs)
                    else:
                        pipe_result_options: TransactionPipeResultOptions = tx.pipe_result_options
                    if pipe_result_options.with_response_options:
                        message = await self.pipe_event_with_response(message, emitter, message_key_as_event,
                                                                      pipe_result_options, **kwargs)
                    else:
                        self.pipe_event(message, emitter, message_key_as_event, pipe_result_options, **kwargs)
        except Exception as e:
            self.logger.error(f'execute => Exception: '
                              f'type: {sys.exc_info()[0]}; '
                              f'line#: {sys.exc_info()[2].tb_lineno}; '
                              f'msg: {str(e)}')

    def pipe_event(
            self,
            message,
            emitter,
            message_key_as_event,
            options: TransactionPipeResultOptions,
            **kwargs
    ):
        event_id = None
        headers = kwargs.get('headers')
        if headers:
            headers = dict((x, y.decode('utf-8')) for x, y in kwargs.get('headers'))
            event_id = headers.get('event_id')
        if not event_id:
            event_id = str(uuid.uuid4())
        if headers:
            headers['event_id'] = event_id.encode('utf-8')
            headers = list(headers.items())
        else:
            headers = [('event_id', event_id.encode('utf-8'))]
        try:
            self.logger.info(
                f'<-------- PIPING EVENT: '
                f'name: {options.pipe_event_name}; '
                f'to: {options.pipe_to_topic}; '
                f'event_id: {event_id}')

            if not message_key_as_event:
                message['event'] = options.pipe_event_name
                emitter(
                    options.pipe_to_topic,
                    ProducerRecord(value=json.dumps(message), headers=headers)
                )
            else:
                emitter(
                    options.pipe_to_topic,
                    ProducerRecord(key=options.pipe_event_name, value=json.dumps(message),
                                   headers=[('event_id', event_id.encode('utf-8'))])
                )
            return event_id
        except Exception as e:
            self.logger.error(f'pipe_event => Exception: '
                              f'type: {sys.exc_info()[0]}; '
                              f'line#: {sys.exc_info()[2].tb_lineno}; '
                              f'msg: {str(e)}')
            return None

    async def pipe_event_with_response(
            self,
            message,
            emitter,
            message_key_as_event,
            options: TransactionPipeResultOptions,
            **kwargs
    ):
        cache_client = options.with_response_options.cache_client
        event_id = self.pipe_event(message, emitter, message_key_as_event, options)

        if not event_id:
            raise RuntimeError('Pipe event failed')

        time_up = time.time()
        while True:
            response = cache_client.get(_get_event_id_hash(event_id, self.app_id))
            if response is not None:
                response_msg = json.loads(response.decode('utf-8'))
                self.logger.info(
                    f'--------> PIPE RESPONSE RECEIVED: '
                    f'name: {options.with_response_options.response_event_name}; '
                    f'event_id: {event_id}')
                return response_msg
            else:
                if time.time() - time_up < options.with_response_options.return_event_timeout:
                    await asyncio.sleep(0.001)
                else:
                    raise TimeoutError(f'Pipe event with response timeout: '
                                       f'piped event: {options.pipe_event_name}; '
                                       f'to: {options.pipe_to_topic}; '
                                       f'response event: {options.with_response_options.response_event_name}; '
                                       f'from: {options.with_response_options.response_from_topic} '
                                       f'event_id: {event_id}')


class EmitWithResponseOptions(pydantic.BaseModel):
    topic_event_list: List[Tuple[str, str]]
    cache_client: Any
    return_event_timeout: int


class AppConfig(pydantic.BaseModel):
    app_name: Optional[str]
    app_id: Optional[str]
    bootstrap_servers: List[str]
    producer_config: Optional[Dict]
    consumer_config: Optional[Dict]
    listen_topics: List[str]
    message_key_as_event: Optional[bool]
    message_value_cls: Optional[Dict[str, Any]]
    middleware_message_cb: Optional[Union[
        Callable[[ConsumerRecord], None],
        Callable[[ConsumerRecord], Coroutine[Any, Any, None]]
    ]]
    emit_with_response_options: Optional[EmitWithResponseOptions]
    pipelines_map: Optional[Dict[str, MessagePipeline]]
    max_concurrent_tasks: Optional[int]
    max_concurrent_pipelines: Optional[int]
    logger: Optional[Any]


class KafkaApp:

    def __init__(self, config: AppConfig):
        self.config = config
        if not self.config.logger:
            logging.basicConfig()
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
        else:
            self.logger = self.config.logger
        if self.config.app_name:
            self.app_name = self.config.app_name
        else:
            self.app_name = 'Kafka application'
        self.producer = KafkaConnector.get_producer(config.bootstrap_servers,
                                                    config.producer_config)
        if self.config.app_id:
            self.app_id = self.config.app_id
        else:
            self.app_id = str(uuid.uuid4())

        kafka_listener_config = ListenerConfig(
            bootstrap_servers=config.bootstrap_servers,
            process_message_cb=self._process_message,
            consumer_config=config.consumer_config,
            topics=config.listen_topics,
            logger=self.logger
        )

        self.listener = KafkaConnector.get_listener(kafka_listener_config)

        self.event_map: Dict = {}
        self.pipelines_map = {}

        if self.config.pipelines_map:
            self.pipelines_map = self.config.pipelines_map
            caching_pipelines_map = {}
            for pipeline in self.pipelines_map.values():
                pipeline.app_id = self.app_id
                for txn in pipeline.transactions:
                    if txn.pipe_result_options and txn.pipe_result_options.with_response_options:
                        self._register_caching_pipeline(
                            (
                                txn.pipe_result_options.with_response_options.response_from_topic,
                                txn.pipe_result_options.with_response_options.response_event_name
                            ),
                            txn.pipe_result_options.with_response_options.cache_client,
                            caching_pipelines_map
                        )
                self.pipelines_map = {**self.pipelines_map, **caching_pipelines_map}
        else:
            self.pipelines_map: Dict[str, MessagePipeline] = {}

        if self.config.emit_with_response_options:
            for topic, event in self.config.emit_with_response_options.topic_event_list:
                self._register_caching_pipeline(
                    (topic, event),
                    self.config.emit_with_response_options.cache_client,
                    self.pipelines_map
                )

        self.sync_tasks_queue = deque()
        self.async_tasks_queue = deque()
        self.pipelines_queue = deque()
        self.caching_queue = deque()

        if self.config.max_concurrent_tasks:
            self.max_concurrent_tasks = self.config.max_concurrent_tasks
        else:
            self.max_concurrent_tasks = 100
        if self.config.max_concurrent_pipelines:
            self.max_concurrent_pipelines = self.config.max_concurrent_pipelines
        else:
            self.max_concurrent_pipelines = 100

        self.stop = False

    def _register_caching_pipeline(
            self,
            topic_event: Tuple[str, str],
            cache_client: Any,
            caching_pipelines_map: Dict[str, MessagePipeline]
    ):
        topic, event = topic_event
        key = '.'.join([topic, event])
        caching_pipelines_map[key] = MessagePipeline(
            transactions=[
                MessageTransaction(
                    fnc=self._cache_pipe_response,
                    args={
                        "cache_client": cache_client,
                    }
                )
            ],
            logger=self.logger,
            app_id=self.app_id
        )

    async def _process_message(self, message: ConsumerRecord) -> None:
        try:
            if self.config.middleware_message_cb:
                if not inspect.isfunction(self.config.middleware_message_cb):
                    raise ValueError('middleware_message_cb must be a function')

                if inspect.iscoroutinefunction(self.config.middleware_message_cb):
                    await self.config.middleware_message_cb(message)
                else:
                    self.config.middleware_message_cb(message)

            _message_value = message.value
            pipeline = None

            # Check if message.key to be used as event name
            if self.config.message_key_as_event:
                if self.pipelines_map.get(message.key):
                    pipeline = self.pipelines_map.get(message.key)
                elif self.pipelines_map.get('.'.join([message.topic, message.key])):
                    pipeline = self.pipelines_map.get('.'.join([message.topic, message.key]))

                handle = self.event_map.get('.'.join([message.topic, message.key]))
            else:
                assert _message_value.get('event') is not None, \
                    '"event" property is missing in message.value object. ' \
                    'Provide "event" property or set "message_key_as_event" option to True ' \
                    'to use message.key as event name.'

                if self.pipelines_map.get(_message_value.get('event')):
                    pipeline = self.pipelines_map.get(_message_value.get('event'))
                elif self.pipelines_map.get('.'.join([message.topic, _message_value.get('event')])):
                    pipeline = self.pipelines_map.get('.'.join([message.topic, _message_value.get('event')]))

                handle = self.event_map.get('.'.join([message.topic, _message_value.get('event')]))

            # Check if dataclass is provided for message.value
            if self.config.message_value_cls:
                _message_value_cls = self.config.message_value_cls[message.topic]
                _message = _message_value_cls(**_message_value)

            else:
                _message = _message_value

            kwargs = {
                "topic": message.topic,
                "partition": message.partition,
                "offset": message.offset,
                "key": message.key,
                "headers": message.headers,
                "timestamp": message.timestamp,
                "timestamp_type": message.timestamp_type,
                "checksum": message.checksum,
                "serialized_key_size": message.serialized_key_size,
                "serialized_value_size": message.serialized_value_size,
                "serialized_header_size": message.serialized_header_size
            }

            if handle:
                if inspect.iscoroutinefunction(handle):
                    self.async_tasks_queue.append((handle, _message, kwargs))
                else:
                    self.sync_tasks_queue.append((handle, _message, kwargs))

            if pipeline:
                if getattr(pipeline.transactions[0].fnc, "__name__") == '_cache_pipe_response':
                    self.caching_queue.append((pipeline, _message_value, kwargs))
                else:
                    self.pipelines_queue.append((pipeline, _message_value, kwargs))

        except Exception as e:
            self.logger.error(f'_process_message => Exception: '
                              f'type: {sys.exc_info()[0]};'
                              f' line#: {sys.exc_info()[2].tb_lineno};'
                              f' msg: {str(e)}')

    async def _cache_pipe_response(
            self,
            message,
            logger,
            cache_client,
            **kwargs
    ):
        try:
            if self.config.message_key_as_event:
                event_name = kwargs.get('key')
            else:
                event_name = message["event"]
            headers = dict((x, y.decode('utf-8')) for x, y in kwargs.get('headers'))

            if logger:
                logger.info(
                    f'--------> CACHING PIPE RESPONSE: name: {event_name}; event_id: {headers["event_id"]}')

            cache_client.set(
                name=_get_event_id_hash(headers['event_id'], self.app_id),
                value=json.dumps(message).encode('utf-8'),
                ex=30
            )
            return message
        except Exception as e:
            err_msg = f'_cache_pipe_response => Exception: type: {sys.exc_info()[0]}; ' \
                      f'line#: {sys.exc_info()[2].tb_lineno}; ' \
                      f'msg: {str(e)}'
            self.logger.error(err_msg)

    def on(self, event: str, topic: Optional[str] = None):
        """
        Maps decorated function to topic.event key.

        If topic is provided, the decorated function is mapped to particular topic.event key that means an event that
        comes from a topic other than the specified will not be processed.
        Otherwise, event will be processed no matter which topic it comes from.
        :param event: Event name.
        :type event: str
        :param topic: Topic name
        :type topic: str
        :return:
        :rtype:
        """

        def decorator(func):
            if topic:
                self.event_map['.'.join([topic, event])] = func
            else:
                for t in self.config.listen_topics:
                    self.event_map['.'.join([t, event])] = func

            def wrapper(*args, **kwargs):
                func(*args, **kwargs)

            return wrapper

        return decorator

    def emit(self, topic: str, message: ProducerRecord):
        self.producer.send(topic,
                           message.value,
                           message.key,
                           message.headers,
                           message.partition,
                           message.timestamp_ms)

    async def emit_with_response(
            self,
            topic: str,
            message: ProducerRecord
    ):
        try:
            if not self.config.emit_with_response_options:
                raise AssertionError('Please provide emit_with_response_options in the application config.')

            event_id = str(uuid.uuid4())
            if message.headers:
                message.headers.append(('event_id', event_id.encode('utf-8')))
            else:
                message.headers = [('event_id', event_id.encode('utf-8'))]

            self.logger.info(
                f'<-------- PERFORMING EMIT WITH RESPONSE: '
                # f'event: {message.key if self.config.message_key_as_event else message.value["event"]}; '
                f'to: {topic}; '
                f'event_id: {event_id}')

            self.emit(topic, message)

            time_up = time.time()
            while True:
                response = self.config.emit_with_response_options.cache_client.get(_get_event_id_hash(event_id, self.app_id))
                if response is not None:
                    return json.loads(response.decode('utf-8'))
                else:
                    if time.time() - time_up < self.config.emit_with_response_options.return_event_timeout:
                        await asyncio.sleep(0.001)
                    else:
                        raise TimeoutError(f'Emit event with response timeout: '
                                           # f'emitted event: {message.key if self.config.message_key_as_event else message.value["event"]}; '
                                           f'to: {topic}; '
                                           f'event_id: {event_id}')
        except Exception as e:
            self.logger.error(f'txn_es_common => Exception: '
                              f'type: {sys.exc_info()[0]}; '
                              f'line#: {sys.exc_info()[2].tb_lineno}; '
                              f'msg: {str(e)}')

    async def process_sync_tasks(self):
        while True:
            if self.stop:
                if len(self.sync_tasks_queue) == 0:
                    break

            if len(self.sync_tasks_queue) > 0:
                handle, _message, kwargs = self.sync_tasks_queue.popleft()
                handle(_message, **kwargs)
            await asyncio.sleep(0.001)

    async def process_async_tasks(self):
        while True:
            if self.stop:
                if len(self.async_tasks_queue) == 0:
                    break

            if len(self.async_tasks_queue) > 0:
                batch_size = min(self.max_concurrent_tasks, len(self.async_tasks_queue))
                batch = [self.async_tasks_queue.popleft() for _ in range(batch_size)]
                tasks = [handle(_message, **kwargs) for handle, _message, kwargs in batch]
                await asyncio.gather(*tasks)
            await asyncio.sleep(0.001)

    async def process_pipelines(self):
        while True:
            if self.stop:
                if len(self.pipelines_queue) == 0:
                    break

            if len(self.pipelines_queue) > 0:
                batch_size = min(self.max_concurrent_pipelines, len(self.pipelines_queue))
                batch = [self.pipelines_queue.popleft() for _ in range(batch_size)]
                tasks = [
                    pipeline.execute(_message_value, self.emit, self.config.message_key_as_event, **kwargs)
                    for pipeline, _message_value, kwargs in batch
                ]
                await asyncio.gather(*tasks)
            await asyncio.sleep(0.001)

    async def process_caching(self):
        while True:
            if self.stop:
                if len(self.caching_queue) == 0:
                    break

            if len(self.caching_queue) > 0:
                pipeline, _message_value, kwargs = self.caching_queue.popleft()
                await pipeline.execute(_message_value, self.emit, self.config.message_key_as_event, **kwargs)
            await asyncio.sleep(0.001)

    async def run(self):
        self.logger.info('{} is up and running.'.format(self.app_name))
        try:
            await asyncio.gather(*[
                self.listener.listen(),
                self.process_sync_tasks(),
                self.process_async_tasks(),
                self.process_pipelines(),
                self.process_caching()
            ])
        except KeyboardInterrupt:
            self.close()

    def close(self):
        self.listener.stop = True
        self.stop = True
        self.producer.flush()
        self.producer.close()
        self.logger.info('{} closed.'.format(self.app_name))



from __future__ import absolute_import

__title__ = 'kafka-python-app'
__author__ = 'Dmitry Amanov'
__copyright__ = 'Copyright 2022 Dmitry Amanov'

# Set default logging handler to avoid "No handler found" warnings.
import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())

from kafka_python_app.app import AppConfig, KafkaApp
from kafka_python_app.connector import ListenerConfig, KafkaConnector, ConsumerRecord, ProducerRecord

__all__ = [
    'AppConfig', 'KafkaApp', 'ListenerConfig', 'KafkaConnector', 'ConsumerRecord', 'ProducerRecord',
]



import enum
import sys
import asyncio
import signal
import json
from typing import Optional, Any
import pydantic
import redis
import random
from dateutil import parser
from unittest import IsolatedAsyncioTestCase

sys.path.append('../')
from kafka_python_app.app import AppConfig, KafkaApp, MessagePipeline, MessageTransaction, \
    TransactionPipeResultOptions, TransactionPipeWithReturnOptions
from kafka_python_app.connector import KafkaConnector
from loguru_logger_lite import Logger, Sink, Sinks, BaseSinkOptions, LogLevels


KILL = False


class GracefulKiller:
    objs = []

    def exit_gracefully(self, *args):
        global KILL
        KILL = True
        if len(self.objs) > 1:
            for obj in self.objs:
                obj.close()


class LogMessage(pydantic.BaseModel):
    service_name: str
    service_id: str
    module: str
    component: str
    pid: int
    level: str
    timestamp: float
    text: str


class Events(enum.Enum):
    PROCESS_PERSON = 'process_person'
    PERSON_ADD_MIDDLE_NAME = 'person_add_middle_name'
    PERSON_MULTIPLY_AGE = 'person_multiply_age'
    PERSON_PROCESSED = 'person_processed'

    PROCESS_COMPANY = 'process_company'
    COMPANY_ADD_INC = 'company_add_inc'
    COMPANY_DOUBLE_STOCK_PRICE = 'company_double_stock_price'
    COMPANY_PROCESSED = 'company_processed'


class Topics(enum.Enum):
    APP_1 = 'test_topic_1'
    APP_2 = 'test_topic_2'
    APP_3 = 'test_topic_3'


class PersonPayload(pydantic.BaseModel):
    first_name: str
    last_name: str
    age: int


class CompanyPayload(pydantic.BaseModel):
    name: str
    stock_value: float


class Message(pydantic.BaseModel):
    event: str
    payload: Any


killer = GracefulKiller()
signal.signal(signal.SIGINT, killer.exit_gracefully)
signal.signal(signal.SIGTERM, killer.exit_gracefully)

msg_count = 300
persons_processed = 0
companies_processed = 0


async def print_result_company(message, logger, **kwarg):
    global companies_processed
    if not message.get('error'):
        logger.info(f'Company processing completed: result: {CompanyPayload(**message["payload"])}')
    else:
        logger.error(f'Company processing completed with error: result: {message["error"]}')
    companies_processed += 1


async def pipe_event(message, logger, **kwargs):
    return message


async def person_add_middle_name(message, logger, **kwargs):
    if not message.get('error'):
        person = PersonPayload(**message['payload'])
        devider = kwargs.get('devider')
        middle_name = kwargs.get('middle_name')
        person.first_name = devider.join(person.first_name.split(devider) + [middle_name])
        message['payload'] = json.loads(person.json(exclude_unset=True))
    await asyncio.sleep(random.randint(1, 5) / 10)
    logger.info(f'Executing transaction: "person_add_middle_name" ==> result: {message}')
    return message


async def person_multiply_age(message, logger, **kwargs):
    if not message.get('error'):
        person = PersonPayload(**message['payload'])
        multiplier = kwargs.get('multiplier')
        person.age = person.age * multiplier
        message['payload'] = json.loads(person.json(exclude_unset=True))
    await asyncio.sleep(random.randint(1, 5) / 10)
    logger.info(f'Executing transaction: "person_multiply_age" ==> result: {message}')
    return message


async def company_add_inc(message, logger, **kwargs):
    if not message.get('error'):
        company = CompanyPayload(**message['payload'])
        company.name = company.name + ' INC.'
        message['payload'] = json.loads(company.json(exclude_unset=True))
    await asyncio.sleep(random.randint(1, 5) / 10)
    logger.info(f'Executing transaction: "company_add_inc" ==> result: {message}')
    return message


async def company_double_stock_price(message, logger: Optional[Any], **kwargs):
    if not message.get('error'):
        company = CompanyPayload(**message['payload'])
        company.stock_value = company.stock_value * 2
        message['payload'] = json.loads(company.json(exclude_unset=True))
    await asyncio.sleep(random.randint(1, 5) / 10)
    logger.info(f'Executing transaction: "company_double_stock_price" ==> result: {message}')
    return message


class TestKafkaApp(IsolatedAsyncioTestCase):
    LOGGER_STDOUT_FMT = "SERVICE NAME: <yellow>{extra[service_name]}</yellow> " \
                        "| SERVICE ID: <green>{extra[service_id]}</green> " \
                        "| MODULE: <green>{module}</green> | COMPONENT: <yellow>{name}</yellow> " \
                        "| PID: {process} | <level>{level}</level> | {time} | <level>{message}</level>"
    LOGGER_PLAIN_FMT = "SERVICE NAME: {extra[service_name]} | SERVICE ID: {extra[service_id]} " \
                       "| MODULE: {module} | COMPONENT: {name} | PID: {process} | {level} | {time} | {message}"

    @staticmethod
    def log_serializer(data):
        formatted = LogMessage(
            service_name=data.split('|')[0].split(':')[1].strip(),
            service_id=data.split('|')[1].split(':')[1].strip(),
            module=data.split('|')[2].split(':')[1].strip(),
            component=data.split('|')[3].split(':')[1].strip(),
            pid=data.split('|')[4].split(':')[1].strip(),
            level=data.split('|')[5].strip(),
            timestamp=parser.parse(data.split('|')[6].strip()).timestamp() * 1000,
            text=data.split('|')[7].strip()
        )
        return formatted.json(exclude_unset=True).encode('utf-8')

    LOGGER_0 = Logger.get_logger([
        Sink(
            name=Sinks.STDOUT,
            opts=BaseSinkOptions(
                level=LogLevels.DEBUG,
                format=LOGGER_STDOUT_FMT,
            )
        ),
    ])
    LOGGER_1 = Logger.get_logger([
        Sink(
            name=Sinks.STDOUT,
            opts=BaseSinkOptions(
                level=LogLevels.DEBUG,
                format=LOGGER_STDOUT_FMT,
            )
        ),
    ])
    LOGGER_2 = Logger.get_logger([
        Sink(
            name=Sinks.STDOUT,
            opts=BaseSinkOptions(
                level=LogLevels.DEBUG,
                format=LOGGER_STDOUT_FMT,
            )
        ),
    ])
    LOGGER_3 = Logger.get_logger([
        Sink(
            name=Sinks.STDOUT,
            opts=BaseSinkOptions(
                level=LogLevels.DEBUG,
                format=LOGGER_STDOUT_FMT,
            )
        ),
    ])

    LOGGER_0 = LOGGER_0.bind(service_name='Tester', service_id='tester')
    LOGGER_1 = LOGGER_1.bind(service_name='App 1', service_id='app_1')
    LOGGER_2 = LOGGER_2.bind(service_name='App 2', service_id='app_2')
    LOGGER_3 = LOGGER_3.bind(service_name='App 3', service_id='app_3')

    KAFKA_BOOTSTRAP_SERVERS = ['127.0.0.1:9092']
    CACHE_SERVER_ADDRESS = '127.0.0.1:6379'
    CACHE_SERVER_PW = 'pass'
    PIPED_EVENT_RETURN_TIMEOUT = 10

    cache_server_ip, cache_server_port = CACHE_SERVER_ADDRESS.split(':')
    event_cache_client = redis.Redis(
        host=cache_server_ip,
        port=int(cache_server_port),
        password=CACHE_SERVER_PW,
        db=0)

    # Scenario 1
    # App_1 receives event 'process_person' (test_topic_1)
    # App_1 pipes event 'person_add_middle_name' to App_2 (test_topic_2)
    # App_2 pipes event 'person_multiply_age' to App_3 (test_topic_3)
    # App_3 pipes event 'person_processed' back to App_1 (test_topic_1)
    # App_1 receives event (standalone event handler) 'person_processed' and prints result

    # Scenario 2
    # App_1 receives event 'process_company'
    # App_1 pipes event with return 'company_add_inc' to App_2
    # App_2 pipes event 'company_add_inc' back to App_1
    # App_1 pipes event with return 'company_double_stock_price' to App_3
    # App_3 pipes event 'company_processed' back to App_1
    # App_1 prints result

    app1_process_person_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=pipe_event,
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.PERSON_ADD_MIDDLE_NAME.value,
                    pipe_to_topic=Topics.APP_2.value
                )
            )
        ],
        logger=LOGGER_1
    )
    app2_process_person_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=person_add_middle_name,
                args={
                    'devider': '-',
                    'middle_name': 'Joe'
                },
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.PERSON_MULTIPLY_AGE.value,
                    pipe_to_topic=Topics.APP_3.value
                )
            )
        ],
        logger=LOGGER_2
    )
    app3_process_person_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=person_multiply_age,
                args={
                    'multiplier': 2
                },
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.PERSON_PROCESSED.value,
                    pipe_to_topic=Topics.APP_1.value
                )
            )
        ],
        logger=LOGGER_3
    )

    app1_process_company_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=pipe_event,
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.COMPANY_ADD_INC.value,
                    pipe_to_topic=Topics.APP_2.value,
                    with_response_options=TransactionPipeWithReturnOptions(
                        response_event_name=Events.COMPANY_ADD_INC.value,
                        response_from_topic=Topics.APP_1.value,
                        cache_client=event_cache_client,
                        return_event_timeout=PIPED_EVENT_RETURN_TIMEOUT
                    )
                )
            ),
            MessageTransaction(
                fnc=pipe_event,
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.COMPANY_DOUBLE_STOCK_PRICE.value,
                    pipe_to_topic=Topics.APP_3.value,
                    with_response_options=TransactionPipeWithReturnOptions(
                        response_event_name=Events.COMPANY_PROCESSED.value,
                        response_from_topic=Topics.APP_1.value,
                        cache_client=event_cache_client,
                        return_event_timeout=PIPED_EVENT_RETURN_TIMEOUT
                    )
                )
            ),
            MessageTransaction(
                fnc=print_result_company
            )
        ],
        logger=LOGGER_1
    )
    app2_process_company_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=company_add_inc,
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.COMPANY_ADD_INC.value,
                    pipe_to_topic=Topics.APP_1.value
                )
            ),
        ],
        logger=LOGGER_2
    )
    app3_process_company_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=company_double_stock_price,
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.COMPANY_PROCESSED.value,
                    pipe_to_topic=Topics.APP_1.value
                )
            ),
        ],
        logger=LOGGER_3
    )

    app1_pipelines_map = {
        Events.PROCESS_PERSON.value: app1_process_person_pipeline,
        Events.PROCESS_COMPANY.value: app1_process_company_pipeline
    }
    app2_pipelines_map = {
        Events.PERSON_ADD_MIDDLE_NAME.value: app2_process_person_pipeline,
        Events.COMPANY_ADD_INC.value: app2_process_company_pipeline
    }
    app3_pipelines_map = {
        Events.PERSON_MULTIPLY_AGE.value: app3_process_person_pipeline,
        Events.COMPANY_DOUBLE_STOCK_PRICE.value: app3_process_company_pipeline
    }

    app1_config = AppConfig(
        app_name='App 1',
        app_id='app_1',
        bootstrap_servers=KAFKA_BOOTSTRAP_SERVERS,
        consumer_config={
            'group_id': 'test_app1_group'
        },
        listen_topics=[Topics.APP_1.value],
        pipelines_map=app1_pipelines_map,
        logger=LOGGER_1
    )
    app2_config = AppConfig(
        app_name='App 2',
        app_id='app_2',
        bootstrap_servers=KAFKA_BOOTSTRAP_SERVERS,
        consumer_config={
            'group_id': 'test_app2_group'
        },
        listen_topics=[Topics.APP_2.value],
        pipelines_map=app2_pipelines_map,
        logger=LOGGER_2
    )
    app3_config = AppConfig(
        app_name='App 3',
        app_id='app_3',
        bootstrap_servers=KAFKA_BOOTSTRAP_SERVERS,
        consumer_config={
            'group_id': 'test_app3_group'
        },
        listen_topics=[Topics.APP_3.value],
        pipelines_map=app3_pipelines_map,
        logger=LOGGER_3
    )

    producer = KafkaConnector.get_producer(KAFKA_BOOTSTRAP_SERVERS)
    app_1 = KafkaApp(app1_config)
    app_2 = KafkaApp(app2_config)
    app_3 = KafkaApp(app3_config)
    killer.objs = [app_1, app_2, app_3]

    async def watch_counter(self):
        while True:
            if persons_processed + companies_processed == msg_count * 2:
                self.LOGGER_0.info('Processing completed. Closing...')
                self.app_1.close()
                self.app_2.close()
                self.app_3.close()
                break
            if KILL:
                break
            await asyncio.sleep(0.001)

    async def test_handle(self):
        @self.app_1.on(Events.PERSON_PROCESSED.value)
        def handle_person(message, **kwargs):
            global persons_processed
            if not message.get('error'):
                self.LOGGER_1.info(f'Person processing completed: result: {PersonPayload(**message["payload"])}')
            else:
                self.LOGGER_1.error(f'Person processing completed with error: result: {message["error"]}')
            persons_processed += 1

        first_names = [
            'John',
            'Frank',
            'Alice',
            'Bob',
            'Julie'
        ]
        last_names = [
            'Doe',
            'Anderson',
            'Ale',
            'Random',
            'First'
        ]
        company_names = [
            'SomeCompany X',
            'SomeCompany Y',
            'SomeCompany Z',
            'SomeCompany J',
            'SomeCompany K'
        ]
        messages_persons = [
            {
                'event': Events.PROCESS_PERSON.value,
                'payload': {
                    'first_name': first_names[random.randint(0, 4)],
                    'last_name': last_names[random.randint(0, 4)],
                    'age': random.randint(15, 70)
                }
            } for i in range(msg_count)
        ]
        messages_company = [
            {
                'event': Events.PROCESS_COMPANY.value,
                'payload': {
                    'name': company_names[random.randint(0, 4)],
                    'stock_value': random.randint(20, 1500)
                }
            } for i in range(msg_count)
        ]
        messages = messages_persons + messages_company

        for msg_obj in messages:
            msg = Message(**msg_obj)
            await asyncio.sleep(0.001)
            self.producer.send(Topics.APP_1.value, json.loads(msg.json(exclude_unset=True)))

        await asyncio.sleep(0.01)
        self.producer.close()
        await asyncio.gather(self.app_1.run(), self.app_2.run(), self.app_3.run(), self.watch_counter())




from setuptools import setup, find_packages
import pathlib

# The directory containing this file
ROOOT = pathlib.Path(__file__).parent

# The text of the README file
README = (ROOOT / "README.md").read_text()

setup(
    name='kafka_python_app',
    version="0.2.8",
    author="Dmitry Amanov",
    author_email="dmitry.amanov@gmail.com",
    description="kafka application endpoint",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/doctor3030/kafka-python-app.git",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "pydantic~=1.8.2",
        "setuptools~=57.0.0",
        "kafka-python~=2.0.2",
    ]
)



import enum
import sys
import asyncio
import signal
import json
# import time
from typing import Optional, Any
import pydantic
import redis
import random
from dateutil import parser
from unittest import IsolatedAsyncioTestCase

sys.path.append('../')
from kafka_python_app.app import AppConfig, KafkaApp, MessagePipeline, MessageTransaction, \
    TransactionPipeResultOptions, EmitWithResponseOptions
from kafka_python_app.connector import KafkaConnector, ProducerRecord
from loguru_logger_lite import Logger, Sink, Sinks, BaseSinkOptions, LogLevels

KILL = False


class GracefulKiller:
    objs = []

    def exit_gracefully(self, *args):
        global KILL
        KILL = True
        if len(self.objs) > 1:
            for obj in self.objs:
                obj.close()


class LogMessage(pydantic.BaseModel):
    service_name: str
    service_id: str
    module: str
    component: str
    pid: int
    level: str
    timestamp: float
    text: str


class Events(enum.Enum):
    PROCESS_PERSON = 'process_person'
    PERSON_ADD_MIDDLE_NAME = 'person_add_middle_name'
    PERSON_MULTIPLY_AGE = 'person_multiply_age'
    PERSON_PROCESSED = 'person_processed'

    PROCESS_COMPANY = 'process_company'
    COMPANY_ADD_INC = 'company_add_inc'
    COMPANY_DOUBLE_STOCK_PRICE = 'company_double_stock_price'
    COMPANY_PROCESSED = 'company_processed'


class Topics(enum.Enum):
    APP_1 = 'test_topic_1'
    APP_2 = 'test_topic_2'
    APP_3 = 'test_topic_3'


class PersonPayload(pydantic.BaseModel):
    first_name: str
    last_name: str
    age: int


class CompanyPayload(pydantic.BaseModel):
    name: str
    stock_value: float


class Message(pydantic.BaseModel):
    event: str
    payload: Any


killer = GracefulKiller()
signal.signal(signal.SIGINT, killer.exit_gracefully)
signal.signal(signal.SIGTERM, killer.exit_gracefully)


async def person_add_middle_name(message, logger, **kwargs):
    if not message.get('error'):
        person = PersonPayload(**message['payload'])
        devider = kwargs.get('devider')
        middle_name = kwargs.get('middle_name')
        person.first_name = devider.join(person.first_name.split(devider) + [middle_name])
        message['payload'] = json.loads(person.json(exclude_unset=True))
    await asyncio.sleep(random.randint(1, 5) / 10)
    logger.info(f'Executing transaction: "person_add_middle_name" ==> result: {message}')
    return message


async def person_multiply_age(message, logger, **kwargs):
    if not message.get('error'):
        person = PersonPayload(**message['payload'])
        multiplier = kwargs.get('multiplier')
        person.age = person.age * multiplier
        message['payload'] = json.loads(person.json(exclude_unset=True))
    await asyncio.sleep(random.randint(1, 5) / 10)
    logger.info(f'Executing transaction: "person_multiply_age" ==> result: {message}')
    return message


async def company_add_inc(message, logger, **kwargs):
    if not message.get('error'):
        company = CompanyPayload(**message['payload'])
        company.name = company.name + ' INC.'
        message['payload'] = json.loads(company.json(exclude_unset=True))
    await asyncio.sleep(random.randint(1, 5) / 10)
    logger.info(f'Executing transaction: "company_add_inc" ==> result: {message}')
    return message


async def company_double_stock_price(message, logger: Optional[Any], **kwargs):
    if not message.get('error'):
        company = CompanyPayload(**message['payload'])
        company.stock_value = company.stock_value * 2
        message['payload'] = json.loads(company.json(exclude_unset=True))
    await asyncio.sleep(random.randint(1, 5) / 10)
    logger.info(f'Executing transaction: "company_double_stock_price" ==> result: {message}')
    return message


class TestKafkaApp(IsolatedAsyncioTestCase):
    LOGGER_STDOUT_FMT = "SERVICE NAME: <yellow>{extra[service_name]}</yellow> " \
                        "| SERVICE ID: <green>{extra[service_id]}</green> " \
                        "| MODULE: <green>{module}</green> | COMPONENT: <yellow>{name}</yellow> " \
                        "| PID: {process} | <level>{level}</level> | {time} | <level>{message}</level>"
    LOGGER_PLAIN_FMT = "SERVICE NAME: {extra[service_name]} | SERVICE ID: {extra[service_id]} " \
                       "| MODULE: {module} | COMPONENT: {name} | PID: {process} | {level} | {time} | {message}"

    @staticmethod
    def log_serializer(data):
        formatted = LogMessage(
            service_name=data.split('|')[0].split(':')[1].strip(),
            service_id=data.split('|')[1].split(':')[1].strip(),
            module=data.split('|')[2].split(':')[1].strip(),
            component=data.split('|')[3].split(':')[1].strip(),
            pid=data.split('|')[4].split(':')[1].strip(),
            level=data.split('|')[5].strip(),
            timestamp=parser.parse(data.split('|')[6].strip()).timestamp() * 1000,
            text=data.split('|')[7].strip()
        )
        return formatted.json(exclude_unset=True).encode('utf-8')

    LOGGER_0 = Logger.get_logger([
        Sink(
            name=Sinks.STDOUT,
            opts=BaseSinkOptions(
                level=LogLevels.DEBUG,
                format=LOGGER_STDOUT_FMT,
            )
        ),
    ])
    LOGGER_1 = Logger.get_logger([
        Sink(
            name=Sinks.STDOUT,
            opts=BaseSinkOptions(
                level=LogLevels.DEBUG,
                format=LOGGER_STDOUT_FMT,
            )
        ),
    ])
    LOGGER_2 = Logger.get_logger([
        Sink(
            name=Sinks.STDOUT,
            opts=BaseSinkOptions(
                level=LogLevels.DEBUG,
                format=LOGGER_STDOUT_FMT,
            )
        ),
    ])
    LOGGER_3 = Logger.get_logger([
        Sink(
            name=Sinks.STDOUT,
            opts=BaseSinkOptions(
                level=LogLevels.DEBUG,
                format=LOGGER_STDOUT_FMT,
            )
        ),
    ])

    LOGGER_0 = LOGGER_0.bind(service_name='Tester', service_id='tester')
    LOGGER_1 = LOGGER_1.bind(service_name='App 1', service_id='app_1')
    LOGGER_2 = LOGGER_2.bind(service_name='App 2', service_id='app_2')
    LOGGER_3 = LOGGER_3.bind(service_name='App 3', service_id='app_3')

    KAFKA_BOOTSTRAP_SERVERS = ['127.0.0.1:9092']
    CACHE_SERVER_ADDRESS = '127.0.0.1:6379'
    CACHE_SERVER_PW = 'pass'
    PIPED_EVENT_RETURN_TIMEOUT = 10

    cache_server_ip, cache_server_port = CACHE_SERVER_ADDRESS.split(':')
    event_cache_client = redis.Redis(
        host=cache_server_ip,
        port=int(cache_server_port),
        password=CACHE_SERVER_PW,
        db=0)

    # Scenario 1
    # App_1 receives event 'process_person' (test_topic_1)
    # App_1 pipes event 'person_add_middle_name' to App_2 (test_topic_2)
    # App_2 pipes event 'person_multiply_age' to App_3 (test_topic_3)
    # App_3 pipes event 'person_processed' back to App_1 (test_topic_1)
    # App_1 receives event (standalone event handler) 'person_processed' and prints result

    # Scenario 2
    # App_1 receives event 'process_company'
    # App_1 pipes event with return 'company_add_inc' to App_2
    # App_2 pipes event 'company_add_inc' back to App_1
    # App_1 pipes event with return 'company_double_stock_price' to App_3
    # App_3 pipes event 'company_processed' back to App_1
    # App_1 prints result

    app2_person_mid_name_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=person_add_middle_name,
                args={
                    'devider': '-',
                    'middle_name': 'Joe'
                },
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.PERSON_ADD_MIDDLE_NAME.value,
                    pipe_to_topic=Topics.APP_1.value
                )
            )
        ],
        logger=LOGGER_2
    )
    app2_person_age_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=person_multiply_age,
                args={
                    'multiplier': 2
                },
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.PERSON_MULTIPLY_AGE.value,
                    pipe_to_topic=Topics.APP_1.value
                )
            )
        ],
        logger=LOGGER_2
    )

    app3_company_add_inc_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=company_add_inc,
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.COMPANY_ADD_INC.value,
                    pipe_to_topic=Topics.APP_1.value
                )
            ),
        ],
        logger=LOGGER_2
    )
    app3_company_stock_price_pipeline = MessagePipeline(
        transactions=[
            MessageTransaction(
                fnc=company_double_stock_price,
                pipe_result_options=TransactionPipeResultOptions(
                    pipe_event_name=Events.COMPANY_DOUBLE_STOCK_PRICE.value,
                    pipe_to_topic=Topics.APP_1.value
                )
            ),
        ],
        logger=LOGGER_3
    )

    app2_pipelines_map = {
        Events.PERSON_ADD_MIDDLE_NAME.value: app2_person_mid_name_pipeline,
        Events.PERSON_MULTIPLY_AGE.value: app2_person_age_pipeline
    }
    app3_pipelines_map = {
        Events.COMPANY_ADD_INC.value: app3_company_add_inc_pipeline,
        Events.COMPANY_DOUBLE_STOCK_PRICE.value: app3_company_stock_price_pipeline
    }

    app1_config = AppConfig(
        app_name='App 1',
        app_id='app_1',
        bootstrap_servers=KAFKA_BOOTSTRAP_SERVERS,
        consumer_config={
            'group_id': 'test_app1_group'
        },
        listen_topics=[Topics.APP_1.value],
        emit_with_response_options=EmitWithResponseOptions(
            topic_event_list=[
                (Topics.APP_1.value, Events.PERSON_ADD_MIDDLE_NAME.value),
                (Topics.APP_1.value, Events.PERSON_MULTIPLY_AGE.value),
                (Topics.APP_1.value, Events.COMPANY_ADD_INC.value),
                (Topics.APP_1.value, Events.COMPANY_DOUBLE_STOCK_PRICE.value),
            ],
            cache_client=event_cache_client,
            return_event_timeout=30
        ),
        logger=LOGGER_1
    )
    app2_config = AppConfig(
        app_name='App 2',
        app_id='app_2',
        bootstrap_servers=KAFKA_BOOTSTRAP_SERVERS,
        consumer_config={
            'group_id': 'test_app2_group'
        },
        listen_topics=[Topics.APP_2.value],
        pipelines_map=app2_pipelines_map,
        logger=LOGGER_2
    )
    app3_config = AppConfig(
        app_name='App 3',
        app_id='app_3',
        bootstrap_servers=KAFKA_BOOTSTRAP_SERVERS,
        consumer_config={
            'group_id': 'test_app3_group'
        },
        listen_topics=[Topics.APP_3.value],
        pipelines_map=app3_pipelines_map,
        logger=LOGGER_3
    )

    producer = KafkaConnector.get_producer(KAFKA_BOOTSTRAP_SERVERS)
    app_1 = KafkaApp(app1_config)
    app_2 = KafkaApp(app2_config)
    app_3 = KafkaApp(app3_config)
    killer.objs = [app_1, app_2, app_3]

    messages = [
        {
            'event': Events.PERSON_ADD_MIDDLE_NAME.value,
            'payload': {
                'first_name': 'John',
                'last_name': 'Doe',
                'age': 35
            }
        },
        {
            'event': Events.PERSON_MULTIPLY_AGE.value,
            'payload': {
                'first_name': 'John',
                'last_name': 'Doe',
                'age': 35
            }
        },
        {
            'event': Events.COMPANY_ADD_INC.value,
            'payload': {
                'name': 'SomeCompany',
                'stock_value': 1224.55
            }
        },
        {
            'event': Events.COMPANY_DOUBLE_STOCK_PRICE.value,
            'payload': {
                'name': 'SomeCompany',
                'stock_value': 1224.55
            }
        },
    ]

    async def calls(self):
        await asyncio.sleep(1)
        res = await self.app_1.emit_with_response(
            Topics.APP_2.value,
            ProducerRecord(value=self.messages[0])
        )
        res = PersonPayload(**res['payload'])

        await asyncio.sleep(1)
        print(await self.app_1.emit_with_response(
            Topics.APP_2.value,
            ProducerRecord(value=self.messages[1])
        ))
        await asyncio.sleep(1)
        print(await self.app_1.emit_with_response(
            Topics.APP_3.value,
            ProducerRecord(value=self.messages[2])
        ))
        await asyncio.sleep(1)
        print(await self.app_1.emit_with_response(
            Topics.APP_3.value,
            ProducerRecord(value=self.messages[3])
        ))
        self.LOGGER_0.info('Processing completed. Closing...')
        # await asyncio.sleep(10)
        self.app_1.close()
        self.app_2.close()
        self.app_3.close()

    async def test(self):
        await asyncio.gather(self.app_1.run(), self.app_2.run(), self.app_3.run(), self.calls())
